/*  Gringotts - a small utility to safe-keep sensitive data
 *  (c) 2002, Germano Rizzo <mano@pluto.linux.it>
 *
 *  grg_fileutils.c - functions to perform various op. on files
 *  Author: Germano Rizzo
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <libgringotts.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>

#include "grg_defs.h"
#include "grg_fileutils.h"

#ifdef HAVE_TCGETATTR
#include <termios.h>
#endif

#define CHAR64(c) \
	((c < 0 || c > 127) ? -1 : index_64[c])

static const char basis_64[] =
  "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

#ifdef HAVE_TCGETATTR
static struct termios termsave;
int outty, fd;
#endif

static const char index_64[128] = {
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
  -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 62, -1, -1, -1, 63,
  52, 53, 54, 55, 56, 57, 58, 59, 60, 61, -1, -1, -1, -1, -1, -1,
  -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14,
  15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, -1, -1, -1, -1, -1,
  -1, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
  41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, -1, -1, -1, -1, -1
};

unsigned char *
encode64 (const unsigned char *in, unsigned int inlen, unsigned int *outlen)
{
  unsigned char *out, *ret;
  unsigned char oval;
  unsigned int olen;

  olen = (inlen + 2) / 3 * 4 + 1;
  out = (unsigned char *) malloc (olen);
  ret = out;
  if (outlen)
    *outlen = olen;

  while (inlen >= 3)
    {
      *out++ = basis_64[in[0] >> 2];
      *out++ = basis_64[((in[0] << 4) & 0x30) | (in[1] >> 4)];
      *out++ = basis_64[((in[1] << 2) & 0x3c) | (in[2] >> 6)];
      *out++ = basis_64[in[2] & 0x3f];
      in += 3;
      inlen -= 3;
    }
  if (inlen > 0)
    {
      *out++ = basis_64[in[0] >> 2];
      oval = (in[0] << 4) & 0x30;
      if (inlen > 1)
	oval |= in[1] >> 4;
      *out++ = basis_64[oval];
      *out++ = (inlen < 2) ? '=' : basis_64[(in[1] << 2) & 0x3c];
      *out++ = '=';
    }

  ret[olen - 1] = '\0';

  return ret;
}

unsigned char *
decode64 (const unsigned char *in, unsigned int inlen, unsigned int *outlen)
{

  unsigned olen, lup;
  int c1, c2, c3, c4;
  char *out, *ret;

  olen = inlen / 4 * 3;
  if (in[inlen - 1] == '=')
    {
      olen--;
      if (in[inlen - 2] == '=')
	olen--;
    }

  out = (char *) malloc (olen);
  ret = out;

  if (in[0] == '+' && in[1] == ' ')
    in += 2;

  if (*in == '\0')
    return NULL;

  for (lup = 0; lup < inlen / 4; lup++)
    {
      c1 = in[0];
      if (CHAR64 (c1) == -1)
	return NULL;
      c2 = in[1];
      if (CHAR64 (c2) == -1)
	return NULL;
      c3 = in[2];
      if (c3 != '=' && CHAR64 (c3) == -1)
	return NULL;
      c4 = in[3];
      if (c4 != '=' && CHAR64 (c4) == -1)
	return NULL;
      in += 4;
      *out++ = (CHAR64 (c1) << 2) | (CHAR64 (c2) >> 4);
      if (c3 != '=')
	{
	  *out++ = ((CHAR64 (c2) << 4) & 0xf0) | (CHAR64 (c3) >> 2);
	  if (c4 != '=')
	    *out++ = ((CHAR64 (c3) << 6) & 0xc0) | CHAR64 (c4);
	}
    }

  if (outlen)
    *outlen = olen;

  return ret;
}

int
file_shred (char *path, int npasses)
{

#define SHRED_BLOCK_SIZE	65536

  int fd, dim, i, j;
  struct stat buf;
  char *mem;
  GRG_CTX gctx;

  fd = open (path, O_RDWR);

  if (fd < 3)
    {
      close (fd);
      return SHRED_ERR_CANT_OPEN_FILE;
    }

  fstat (fd, &buf);

  if (buf.st_nlink > 1)
    {
      close (fd);
      return SHRED_ERR_YET_LINKED;
    }

  dim = buf.st_size;

  mem = mmap (NULL, dim, PROT_WRITE, MAP_SHARED, fd, 0);

  if (mem == MAP_FAILED)
    {
      close (fd);
      return SHRED_ERR_CANT_MMAP;
    }

  gctx = grg_context_initialize_defaults ("GRG");

  for (i = 0; i < npasses; i++)
    {
      int rem = dim % SHRED_BLOCK_SIZE;
      char *rnd;
      for (j = SHRED_BLOCK_SIZE; j < dim; j += SHRED_BLOCK_SIZE)
	{
	  rnd = grg_rnd_seq (gctx, SHRED_BLOCK_SIZE);
	  memcpy (mem + j - SHRED_BLOCK_SIZE, rnd, SHRED_BLOCK_SIZE);
	  free (rnd);
	}
      rnd = grg_rnd_seq (gctx, rem);
      memcpy (mem + dim - rem, rnd, rem);
      free (rnd);
      fsync (fd);
    }

  munmap (mem, dim);
  close (fd);
  unlink (path);
  grg_context_free (gctx);

  sync ();

  return 0;
}

/* these lock and unlock the console, to hide typed chars
the code is taken by GnuPG utils/ttyio.c; in this simple
implementation, it's important not to nest them 
The code has been adapted from the GnuPG sources.
Copyright (C) 2002 Free Software Foundation, Inc.
Thanks to them, and thanks to Free Software! :)
*/

void
block_term (void)
{
#ifdef HAVE_TCGETATTR
	struct termios term;
  #ifdef HAVE_ISATTY
	outty = isatty (STDOUT);
  #else
	outty = 0;
  #endif
	if (!outty)
		fd = open ("/dev/tty", O_RDWR);
	else
		fd = STDOUT;
	tcgetattr(fd, &termsave);
	term = termsave;
	term.c_lflag &= ~(ECHO | ECHOE | ECHOK | ECHONL);
	tcsetattr(fd, TCSAFLUSH, &term );
#endif
}

void
unblock_term (void)
{
#ifdef HAVE_TCGETATTR
	tcsetattr(fd, TCSAFLUSH, &termsave);
	if (!outty)
		close (fd);
#endif
}
