/*  Gringotts - a small utility to safe-keep sensitive data
 *  (c) 2002, Germano Rizzo <mano@pluto.linux.it>
 *
 *  grg_popt.c - commandline argument parsing & task execution
 *  Author: Germano Rizzo
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <fcntl.h>

#include <glib.h>
#include <popt.h>

#include "grg_defs.h"
#include "gringotts.h"
#include "grg_pwd.h"
#include "grg_fileutils.h"

#include "libgringotts.h"

static void
dump_content (gchar *fname, gint ennum, gchar *enpage) {
  GRG_KEY key;
  glong err, len;
  guchar *txt;

#ifndef HAVE_TCGETATTR
  fprintf (stderr, "%s: %s\n", _("Warning"),
	_("it isn't possible to hide password typing; be extremely careful!"));
#endif

  if (!g_file_test (fname, G_FILE_TEST_EXISTS))
    report_err (_("The selected file doesn't exists"), 0, 1, NULL);

  if (g_file_test (fname, G_FILE_TEST_IS_DIR | G_FILE_TEST_IS_SYMLINK))
    report_err (_("You've selected a directory or a symlink"), 0, 1, NULL);

  err = grg_validate_file (gctx, fname);

  switch (err)
    {
    case GRG_OK:
      break;

    case GRG_READ_MAGIC_ERR:
    case GRG_READ_UNSUPPORTED_VERSION:
		report_err (_("This file doesn't seem to be a valid Gringotts one!"), 0,
			1, NULL);

    case GRG_READ_FILE_ERR:
		report_err (_("Uh-oh! I can't read from the file!"), 0, 1, NULL);

    case GRG_READ_CRC_ERR:
    case GRG_READ_COMP_ERR:
    	report_err (_("The file appears to be corrupted! Oh, God!"), 0, 1, NULL);

    case GRG_READ_TOO_BIG_ERR:
    	report_err (_("File is too big"), 0, 1, NULL);

    default:
	  if (err < 0)
		report_err (_("The file appears to be corrupted! Oh, God!"), 0, 1, NULL);
    }

  while (TRUE)
  {
    key = grg_get_cmdline_key ();

    if (!key) {
	  report_err (_("You must enter a valid password!"), 0, 0, NULL);
	  continue;
    }

    err = grg_decrypt_file (gctx, key, fname, &txt, &len);

    grg_key_free (gctx, key);

    switch (err)
	{
	case GRG_OK:
		break;

	case GRG_READ_PWD_ERR:
		report_err (_("Wrong password! Re-enter it"), 0, 0, NULL);
		continue;

	case GRG_READ_ENC_INIT_ERR:
		report_err (_("Problem with libmcrypt, probably a faulty installation"),
			0, 1, NULL);

	//just to be sure...
	default:
	  if (err < 0)
		report_err (_("The file appears to be corrupted! Oh, God!"), 0, 1, NULL);
	}
	
	if (!g_utf8_validate (txt, len, NULL)) {
		GRGFREE (txt, len);
		report_err (_("Saved data contain invalid UTF-8 chars"), 0, 1, NULL);
	}

	break;
  }

	grg_entries_load_from_string (txt, NULL, FALSE);
	GRGFREE (txt, len);

	grg_entries_print (ennum, enpage);
	grg_entries_free ();
}

static void
exit_freeing_ctx (gint code)
{
  grg_context_free (gctx);
  exit (code);
}

void
grg_parse_argv (gint argc, gchar * argv[], gchar ** filename,
		gboolean * rootCheck)
{
  poptContext optCon;
  gchar *wipe, *etit;
  gint passes, ennum;
  gboolean dump, help, strongRnd;

  struct poptOption optionsTable[] = {
    {"help", 'h', POPT_ARG_NONE, &help, 1, _("shows the help"), NULL},
    {"input-file", 'f', POPT_ARG_STRING, filename, 0,
     _("specify the input file to open"), _("FILE")},
#ifndef ROOT_FILTER
    {"allow-root", 's', POPT_ARG_NONE, rootCheck, 1,
     _("allow usage as root -- DANGEROUS"), NULL},
#endif
	{"strong-random", 'R', POPT_ARG_NONE, &strongRnd, 0,
     _("force use of /dev/random -- slower"), NULL},
	{"dump", 'd', POPT_ARG_NONE, &dump, 0,
     _("dump the content of a file"), NULL},
	{"entry-num", 0, POPT_ARG_INT, &ennum, 0,
     _("index of the entry to dump"), _("NUM")},
    {"entry-title", 0, POPT_ARG_STRING, &etit, 0,
     _("title of the entry to dump"), _("TXT")},
    {"wipe-file", 'w', POPT_ARG_STRING, &wipe, 0,
     _("securely wipe a file"), _("FILE")},
	{"wipe-passes", 0, POPT_ARG_INT, &passes, 0,
     _("passes in file wiping"), _("NUM")},
    {NULL, 0, 0, NULL, 0}
  };

  *filename = NULL;
  wipe = NULL;
  etit = NULL;
  passes = 0;
  ennum = -1;
  *rootCheck = FALSE;
  strongRnd = FALSE;
  dump = FALSE;
  help = FALSE;

  optCon = poptGetContext (NULL, argc, (const char **) argv, optionsTable, 0);

  while (poptGetNextOpt (optCon) >= 0);

  if (help)
    {
      poptPrintHelp (optCon, stdout, 0);
	  poptFreeContext (optCon);
	  exit_freeing_ctx (1);
    }

  poptFreeContext (optCon);

  if (strongRnd)
	grg_ctx_set_security_lvl (gctx, GRG_SEC_PARANOIA);

/*quite cerebrotic, I know. The idea is: to ensure that stdin isn't exploitable,
  the best way is to close and reopen it, so that any "abnormal" setting is
  wasted. Since it isn't possible to do this atomically, I check that the former
  file descriptor hasn't been opened in the while by someone that could pretend
  to "be" stdin.*/
  if (!(wipe || dump))
    {
	  int canary;

	  close (STDIN);
	  canary = open ("/dev/null", O_RDONLY);
	  if (canary != STDIN)
		  exit_freeing_ctx (1);
	  return;
	}

//wipe and dump operations are processed without returning to main()

//FIXME this should be in grg_safe.c
#ifdef HAVE_ISATTY
  if (!isatty (STDIN))
	{
	  g_critical ("%s", _("It isn't possible to redirect data to stdin, as it is a potential security flaw."));
	  exit_freeing_ctx (1);
    }
#endif

  if (!grg_security_filter (*rootCheck))
	exit_freeing_ctx (1);

  if (wipe)
	{
		gint res;
		gchar c;
		
		if (!g_file_test (wipe, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_REGULAR))
			report_err (_("The file does not exist"), 0, 1, NULL);
		
		printf ("%s (y/n)", _("Are you sure you want to wipe this file?\n"
			"Its content will be securely erased, so no\n"
			"recover is possible."));

		if (passes < 0 || passes > 32)
			passes = 8;

		c = getchar();

		if (c != 'y' && c != 'Y')
		  exit_freeing_ctx (0);

		res = file_shred (wipe, passes);

		if (res < 0)
			report_err (_("File wiping failed"), 0, 1, NULL);
		
		exit_freeing_ctx (0);
	}

  if (dump)
	{
		if (!*filename)
			report_err (_("You must specify a file to dump (with the -f switch)"), 0, 1, NULL);
		dump_content (*filename, ennum, etit);
		exit_freeing_ctx (0);
	}
}
